@@ -27,6 +27,9 @@ import pytest
 # For accessing attributes of functions.
 import inspect
 
+# For performing deep copies.
+import copy
+
 
 
 # For general array handling.
@@ -415,6 +418,38 @@ def generate_standard_coord_transform_params_4_ctor_params():
 
 
 
+def generate_standard_coord_transform_params_5_ctor_params():
+    center = (100, 0.5)
+    
+    quadratic_radial_distortion_amplitude = -0.5
+
+    elliptical_distortion_vector = (0, 0)
+
+    spiral_distortion_amplitude = 0
+
+    amplitude = 0
+    phase = 0
+    parabolic_distortion_vector = (amplitude*np.cos(phase), 
+                                   amplitude*np.sin(phase))
+    
+    standard_coord_transform_params_ctor_params = \
+        {"center": \
+         center,
+         "quadratic_radial_distortion_amplitude": \
+         quadratic_radial_distortion_amplitude,
+         "elliptical_distortion_vector": \
+         elliptical_distortion_vector,
+         "spiral_distortion_amplitude": \
+         spiral_distortion_amplitude,
+         "parabolic_distortion_vector": \
+         parabolic_distortion_vector, 
+         "skip_validation_and_conversion": \
+         False}
+
+    return standard_coord_transform_params_ctor_params
+
+
+
 def generate_center(instance_idx):
     if instance_idx <= 3:
         func_name = ("generate_standard_coord_transform_params_{}_"
@@ -801,6 +836,8 @@ def test_1_of_DistortionModel():
 
     distortion_model = cls_alias()
 
+    distortion_model = copy.deepcopy(distortion_model)
+
     assert distortion_model.is_azimuthally_symmetric
     assert distortion_model.is_trivial
 
@@ -847,9 +884,12 @@ def test_2_of_DistortionModel():
     for multi_dim_slice_1, multi_dim_slice_2 in zip_obj:
         undistorted_images = generate_undistorted_images()[multi_dim_slice_1]
 
-        method_alias = distortion_model.distort_then_resample_images
-        kwargs = {"undistorted_images": undistorted_images}
-        distorted_then_resampled_images = method_alias(**kwargs)
+        for iteration_idx in range(2):
+            if iteration_idx == 1:
+                distortion_model = copy.deepcopy(distortion_model)
+            method_alias = distortion_model.distort_then_resample_images
+            kwargs = {"undistorted_images": undistorted_images}
+            distorted_then_resampled_images = method_alias(**kwargs)
         
         distorted_then_resampled_images = \
             distorted_then_resampled_images.cpu().detach().numpy()
@@ -929,9 +969,12 @@ def test_3_of_DistortionModel():
     for multi_dim_slice_1, multi_dim_slice_2 in zip_obj:
         distorted_images = generate_distorted_images()[multi_dim_slice_1]
 
-        method_alias = distortion_model.undistort_then_resample_images
-        kwargs = {"distorted_images": distorted_images}
-        undistorted_then_resampled_images = method_alias(**kwargs)
+        for iteration_idx in range(2):
+            if iteration_idx == 1:
+                distortion_model = copy.deepcopy(distortion_model)
+            method_alias = distortion_model.undistort_then_resample_images
+            kwargs = {"distorted_images": distorted_images}
+            undistorted_then_resampled_images = method_alias(**kwargs)
         
         undistorted_then_resampled_images = \
             undistorted_then_resampled_images.cpu().detach().numpy()
@@ -1002,7 +1045,9 @@ def test_5_of_DistortionModel():
                   "convergence_map_of_distorted_then_resampled_images",
                   "mask_frame_of_distorted_then_resampled_images",
                   "flow_field_of_coord_transform", 
-                  "flow_field_of_coord_transform_right_inverse")
+                  "flow_field_of_coord_transform_right_inverse",
+                  "out_of_bounds_map_of_distorted_then_resampled_images",
+                  "out_of_bounds_map_of_undistorted_then_resampled_images")
     for attr_name in attr_names:
         kwargs = distortion_model_ctor_params
         distortion_model = distoptica.DistortionModel(**kwargs)
@@ -1098,7 +1143,9 @@ def test_7_of_DistortionModel():
     attr_names = ("sampling_grid", 
                   "convergence_map_of_distorted_then_resampled_images",
                   "flow_field_of_coord_transform", 
-                  "flow_field_of_coord_transform_right_inverse")
+                  "flow_field_of_coord_transform_right_inverse",
+                  "out_of_bounds_map_of_distorted_then_resampled_images",
+                  "out_of_bounds_map_of_undistorted_then_resampled_images")
     for attr_name in attr_names:
         kwargs = distortion_model_ctor_params
         distortion_model = distoptica.DistortionModel(**kwargs)
@@ -1224,6 +1271,117 @@ def test_3_of_generate_standard_distortion_model():
 
 
 
+def test_4_of_generate_standard_distortion_model():
+    kwargs = \
+        generate_standard_coord_transform_params_5_ctor_params()
+    standard_coord_transform_params = \
+        distoptica.StandardCoordTransformParams(**kwargs)
+    
+    distortion_model_ctor_params = \
+        generate_distortion_model_ctor_params(instance_idx=4)
+
+    kwargs = distortion_model_ctor_params.copy()
+    del kwargs["coord_transform_params"]
+    del kwargs["skip_validation_and_conversion"]
+    kwargs["standard_coord_transform_params"] = standard_coord_transform_params
+    distortion_model = distoptica.generate_standard_distortion_model(**kwargs)
+
+    L, R, B, T = \
+        distortion_model.mask_frame_of_distorted_then_resampled_images
+    sampling_grid_dims_in_pixels = \
+        distortion_model.core_attrs["sampling_grid_dims_in_pixels"]
+
+    assert (L == sampling_grid_dims_in_pixels[0])
+    assert (R == 0)
+    assert (B == 0)
+    assert (T == sampling_grid_dims_in_pixels[1])
+
+    return None
+
+
+
+def test_1_of_apply_coord_transform():
+    q_x, q_y = distoptica.apply_coord_transform()
+
+    kwargs = {"instance_idx": 0}
+    coord_transform_params = generate_coord_transform_params(**kwargs)
+
+    kwargs = {"u_x": q_x,
+              "u_y": q_y,
+              "coord_transform_params": coord_transform_params,
+              "device": q_x.device,
+              "skip_validation_and_conversion": True}
+    _ = distoptica.apply_coord_transform(**kwargs)
+
+    kwargs = \
+        generate_standard_coord_transform_params_1_ctor_params()
+    standard_coord_transform_params = \
+        distoptica.StandardCoordTransformParams(**kwargs)
+
+    kwargs = {"coord_transform_params": standard_coord_transform_params}
+    _ = distoptica.apply_coord_transform(**kwargs)
+
+    with pytest.raises(ValueError) as err_info:
+        kwargs["u_x"] = [[0.5]]
+        kwargs["u_y"] = [[0.5, 0.5]]
+        _ = distoptica.apply_coord_transform(**kwargs)
+
+    kwargs["u_x"] = q_x
+    kwargs["u_y"] = q_y
+    kwargs["device"] = q_x.device
+    _ = distoptica.apply_coord_transform(**kwargs)
+
+    with pytest.raises(TypeError) as err_info:
+        kwargs["u_x"] = torch.tensor([[[0.5]]])
+        kwargs["u_y"] = kwargs["u_x"]
+        _ = distoptica.apply_coord_transform(**kwargs)
+
+    return None
+
+
+
+def test_1_of_apply_coord_transform_right_inverse():
+    _ = distoptica.apply_coord_transform_right_inverse()
+
+    kwargs = {"instance_idx": 0}
+    coord_transform_params = generate_coord_transform_params(**kwargs)
+
+    least_squares_alg_params = generate_least_squares_alg_params()
+
+    expected_q_x = torch.tensor([[0.35, 0.50, 0.65],
+                                 [0.35, 0.50, 0.65],
+                                 [0.35, 0.50, 0.65]])
+    expected_q_y = torch.tensor([[0.35, 0.35, 0.35],
+                                 [0.50, 0.50, 0.50],
+                                 [0.65, 0.65, 0.65]])
+
+    kwargs = \
+        {"q_x": expected_q_x,
+         "q_y": expected_q_y,
+         "coord_transform_params": coord_transform_params,
+         "device": expected_q_x.device,
+         "least_squares_alg_params": least_squares_alg_params,
+         "skip_validation_and_conversion": True}
+    u_x, u_y, convergence_map = \
+        distoptica.apply_coord_transform_right_inverse(**kwargs)
+
+    kwargs = {"u_x": u_x,
+              "u_y": u_y,
+              "coord_transform_params": coord_transform_params,
+              "device": u_x.device,
+              "skip_validation_and_conversion": True}
+    q_x, q_y = distoptica.apply_coord_transform(**kwargs)
+
+    tol = 1e-3
+
+    assert torch.all(convergence_map)
+    assert (torch.all(torch.abs(q_x-expected_q_x) < tol))
+    assert (torch.all(torch.abs(q_y-expected_q_y) < tol))
+
+    return None
+
+
+
 ###########################
 ## Define error messages ##
 ###########################
